////// Nformant - an objects that uses fiddle~ peak data to map the first and second formants to an x-y plane, and the fundamental frequency to a seperate z-axis
////// by Notto J. Thelle, University of Oslo, Norway 2008

// inputs and outputs: input is a mono audio signal, outputs are the first and second formants. 
// These outputs need to be calibrated by the user as explained in the documentation.
inlets = 1;
outlets = 3;

// fiddle~ sends out successive lists consisting of 3 elements: index number, peak frequency and amplitude.
// Depending on how many peaks one chooses to output from the far right outlet of fiddle~ (defined by the fourth 
// argument in the fiddle~ object), the index number (a) grows until it reaches the highest peak number and then starts
// sending out information from the next window. The frequency and amplitude from these lists (b and c) are saved into seperate
// arrays (freq and amp) and given an index number that corresponds with the index number from fiddle~. When the index number
// reaches 20.0, the values from the arrays freq and amp are treated in the function mytask(), the script outputs the formants and the fundamental pitch.
// This process repeats itself, while the Nformant object is completely in sync with fiddle~. If a different amount of peaks to be 
// treated is desirable, make sure that both the fourth argument in the fiddle~ object and the corresponding code matches.
// The number in the code line "if (a == 20.0) {" should be the same as the fourth argument in the fiddle~ object.

// global variables
var freq = new Array();
var amp = new Array();
var tsk = new Task(mytask);
var lastf1;
var lastf2;

// reads all peaks from a fiddle~ window and saves them into the arrays freq and amp
function msg_float(a, b, c) {
    var i;
    for (i = 0; i < 1; i++) {
        freq[a] = b;
        amp[a] = c;
        if (a == 20.0) {
            tsk.execute(); // when index number 20.0 is reached (a full window from fiddle~), the function mytask() is called upon to treat the data from the arrays
        }
    }
}	

function mytask() {
    var frontback = 0.0;
    var highlow = 0.0;
    var f1 = 0.0; // the final values
    var f2 = 0.0; // to be outputted
    var f1freqsum = 0.0; // adds the selected weighted peaks in the first formant frequency range
    var f2freqsum = 0.0; // adds the selected weighted peaks in the second formant frequency range
    var f1ampsum = 0.0; // adds the corresponding weights (amplitudes) from the first formant frequency range
    var f2ampsum = 0.0; // adds the corresponding weights from the second formant frequency range
    var fundamental = freq[1];
    outlet(2, fundamental);
    var f1max = 0.0;
    var f2max = 0.0;
    var f1maxindex = 0;
    var f2maxindex = 0;
    var cpweight = 0.0;
    var k;
    // filters out unwanted data, such as zero values and values outside the f1 and f2 frequency ranges
    for (k = 1; k < freq.length && freq[k] != 0.0 && freq[k] < 2400.0; k++) {
        // below a threshold of 150 Hz, the fundamental is not treated as a formant qualifying peak
        if (fundamental < 150.0 && freq[k] != fundamental) {
            if (freq[k] < 800.00) {
                if (amp[k] > f1max) {
                    f1max = amp[k];
                    f1maxindex = k;
                }
            }
            if (freq[k] >= 800.0) {
                if (freq[f1maxindex] > 400.0) {
                    var f1maxlocation = 800.0 - freq[f1maxindex];
                    var needtoreducef2bottomend = f1maxlocation / 400.0;
                    if (freq[k] < 1000) {
                        var reducedpeak = amp[k] * needtoreducef2bottomend;
                        if (reducedpeak > f2max) {
                            f2max = reducedpeak;
                            f2maxindex = k;
                        }
                    }
                    if (freq[k] >= 1000.0) {
                        if (amp[k] > f2max) {
                            f2max = amp[k];
                            f2maxindex = k;
                        }
                    }
                }
                if (freq[f1maxindex] <= 400.0) {     
                    if (amp[k] > f2max) {
                        f2max = amp[k];
                        f2maxindex = k;
                    }
                }
            }
        }
        // if the fundamental is above 150 Hz, it starts to count as a compromised first formant peak.
        // The fundamental gradually gains relevance, and only
        // fully qualifies as a first formant peak at 300 Hz. This is to avoid having the fundamental tip the scale completely.
        // This script is optimalized for male voices with pitches up to 250 Hz. Above this, the first formant will
        // gradually behave more eratically.
        if (fundamental >= 150.0) {
            if (freq[k] < 800.00) {
                if (freq[k] < 300.0) {
                    var fadeinlocation = freq[k] - 150.0;
                    var fadeinweight = fadeinlocation / 150.0;
                    cpweight = amp[k] * fadeinweight;
                    if (cpweight > f1max) {
                        f1max = cpweight;
                        f1maxindex = k;
                    }
                }
                if (freq[k] >= 300.0 && freq[k] < 800.0) {
                    if (amp[k] > f1max) {
                        f1max = amp[k];
                        f1maxindex = k;
                    }
                }
            }
            if (freq[k] >= 800.0) {
                if (freq[f1maxindex] > 400.0) {
                    var f1maxlocation = 800.0 - freq[f1maxindex];
                    var needtoreducef2bottomend = f1maxlocation / 400.0;
                    if (freq[k] < 1000) {
                        var reducedpeak = amp[k] * needtoreducef2bottomend;
                        if (reducedpeak > f2max) {
                            f2max = reducedpeak;
                            f2maxindex = k;
                        }
                    }
                    if (freq[k] >= 1000.0) {
                        if (amp[k] > f2max) {
                            f2max = amp[k];
                            f2maxindex = k;
                        }
                    }
                }
                if (freq[f1maxindex] <= 400.0) {     
                    if (amp[k] > f2max) {
                        f2max = amp[k];
                        f2maxindex = k;
                    }
                }
            }
        }
    }
    // after having determined which peak index is the central formant peak, the script tests whether the neighbouring peak indexes are
    // "true neighbours". If they are further than a fundamental away, they are ignored. The central formant peak is only weighed against
    // a "true neighbour".
    if (f1maxindex == 1.0) {
        if (freq[f1maxindex + 1] > (fundamental * 2) + 30.0) {
            f1freqsum = freq[f1maxindex] * cpweight;
            f1ampsum = cpweight;
        }
        if (freq[f1maxindex + 1] <= (fundamental * 2) + 30.0) {
            f1freqsum = (freq[f1maxindex] * cpweight) + (freq[f1maxindex + 1] * amp[f1maxindex + 1]);
            f1ampsum = cpweight + amp[f1maxindex + 1];
        }
    }
    if (f1maxindex > 1.0) {
        if (freq[f1maxindex - 1] > (freq[f1maxindex] - fundamental - 30.0) && freq[f1maxindex + 1] < (freq[f1maxindex] + fundamental + 30.0)) {
            f1freqsum = (freq[f1maxindex - 1] * amp[f1maxindex -1]) + (freq[f1maxindex] * amp[f1maxindex]) + (freq[f1maxindex + 1] * amp[f1maxindex + 1]);
            f1ampsum = amp[f1maxindex -1] + amp[f1maxindex] + amp[f1maxindex + 1];
        }
        if (freq[f1maxindex -1] <= (freq[f1maxindex] - fundamental - 30.0) && freq[f1maxindex + 1] < (freq[f1maxindex] + fundamental + 30.0)) {
            f1freqsum = (freq[f1maxindex] * amp[f1maxindex]) + (freq[f1maxindex + 1] * amp[f1maxindex + 1]);
            f1ampsum = amp[f1maxindex] + amp[f1maxindex + 1];
        }
        if (freq[f1maxindex - 1] > (freq[f1maxindex] - fundamental - 30.0) && freq[f1maxindex + 1] >= (freq[f1maxindex] + fundamental + 30.0)) {
            f1freqsum = (freq[f1maxindex - 1] * amp[f1maxindex - 1]) + (freq[f1maxindex] * amp[f1maxindex]);
            f1ampsum = amp[f1maxindex - 1] + amp[f1maxindex];
        }
        if (freq[f1maxindex - 1] <= (freq[f1maxindex] - fundamental - 30.0) && freq[f1maxindex + 1] >= (freq[f1maxindex] + fundamental + 30.0)) {
            f1freqsum = freq[f1maxindex] * amp[f1maxindex];
            f1ampsum = amp[f1maxindex];
        }
    }
    if ((freq[f2maxindex] + fundamental) >= 2400.0) {
        if (freq[f2maxindex - 1] > (freq[f2maxindex] - fundamental - 30.0)) {
            f2freqsum = (freq[f2maxindex - 1] * amp[f2maxindex - 1]) + (freq[f2maxindex] * amp[f2maxindex]);
            f2ampsum = amp[f2maxindex - 1] + amp[f2maxindex];
        }
        if (freq[f2maxindex - 1] <= (freq[f2maxindex] - fundamental - 30.0)) {
            f2freqsum = freq[f2maxindex] * amp[f2maxindex];
        }
    }
    if ((freq[f2maxindex] + fundamental) < 2400.0) {
        if (freq[f2maxindex - 1] > (freq[f2maxindex] - fundamental - 30.0) && freq[f2maxindex + 1] < (freq[f2maxindex] + fundamental + 30.0)) {
            f2freqsum = (freq[f2maxindex - 1] * amp[f2maxindex -1]) + (freq[f2maxindex] * amp[f2maxindex]) + (freq[f2maxindex + 1] * amp[f2maxindex + 1]);
            f2ampsum = amp[f2maxindex -1] + amp[f2maxindex] + amp[f2maxindex + 1];
        }
        if (freq[f2maxindex -1] <= (freq[f2maxindex] - fundamental - 30.0) && freq[f2maxindex + 1] < (freq[f2maxindex] + fundamental + 30.0)) {        
            f2freqsum = (freq[f2maxindex] * amp[f2maxindex]) + (freq[f2maxindex + 1] * amp[f2maxindex + 1]);
            f2ampsum = amp[f2maxindex] + amp[f2maxindex + 1];
        }
        if (freq[f2maxindex - 1] > (freq[f2maxindex] - fundamental - 30.0) && freq[f2maxindex + 1] >= (freq[f2maxindex] + fundamental + 30.0)) {
            f2freqsum = (freq[f2maxindex - 1] * amp[f2maxindex - 1]) + (freq[f2maxindex] * amp[f2maxindex]);
            f2ampsum = amp[f2maxindex - 1] + amp[f2maxindex];
        }
        if (freq[f2maxindex - 1] <= (freq[f2maxindex] - fundamental - 30.0) && freq[f2maxindex + 1] >= (freq[f2maxindex] + fundamental + 30.0)) {
            f2freqsum = freq[f2maxindex] * amp[f2maxindex];
            f2ampsum = amp[f2maxindex];
        }
    }
    f1 = f1freqsum / f1ampsum;
    f2 = f2freqsum / f2ampsum;
    outlet(0, f1);
    outlet(1, f2);
    lastf1 = f1;
    lastf2 = f2;
}